
//Mersenne Twister pseudorandom number generator of 32 and 64-bit.
//2005, Diego Park <diegopark@gmail.com>

#define WIN32_LEAN_AND_MEAN
#include <stdlib.h>
#include <stdio.h>
#include <tchar.h>
#include <windows.h>

#include "mt32.h"
#include "mt64.h"
#include "PkRandom.h"

using namespace System;
using namespace Pk;

static unsigned long dummy;
static unsigned long long dummy64;

int _tmain(int argc, _TCHAR* argv[]) {
	__int64 startRand, endRand, startOriginal, endOriginal, start, end, frequency;

	srand(5489);

	init_genrand(5489);
	init_genrand64(5489);

	Random32 mersenneTwister32(5489);
	Random64 mersenneTwister64(5489);

	const size_t MaxIterations = 100000000;

	if (QueryPerformanceFrequency((LARGE_INTEGER*)&frequency) != FALSE) {
		// rand()
		QueryPerformanceCounter((LARGE_INTEGER*)&startRand);
		for (size_t i = MaxIterations; i > 0; i--) {
			dummy = rand();
		}
		QueryPerformanceCounter((LARGE_INTEGER*)&endRand);

		Console::WriteLine("\n----- Rand -----");
		Console::WriteLine("Time: {0} seconds.", ((endRand - startRand) * 1.0 / frequency).ToString());

		// Original
		QueryPerformanceCounter((LARGE_INTEGER*)&startOriginal);
		for (size_t i = MaxIterations; i > 0; i--) {
			dummy64 = genrand64_int64();
		}
		QueryPerformanceCounter((LARGE_INTEGER*)&endOriginal);

		// Implementation
		QueryPerformanceCounter((LARGE_INTEGER*)&start);
		for (size_t i = MaxIterations; i > 0; i--) {
			dummy64 = mersenneTwister64.next();
		}
		QueryPerformanceCounter((LARGE_INTEGER*)&end);

		Console::WriteLine("\n----- 64-bit test -----");
		Console::WriteLine("Time: {0} seconds.", ((end - start) * 1.0 / frequency).ToString());
		Console::WriteLine("Original time: {0} seconds.", ((endOriginal - startOriginal) * 1.0 / frequency).ToString());
		Console::WriteLine("Speed-up to rand: {0} %.", (((endRand - startRand) * 1.0 / (end - start) * 1.0) * 100.0).ToString());
		Console::WriteLine("Speed-up to original: {0} %.", (((endOriginal - startOriginal) * 1.0 / (end - start) * 1.0) * 100.0).ToString());



		// Original
		QueryPerformanceCounter((LARGE_INTEGER*)&startOriginal);
		for (size_t i = MaxIterations; i > 0; i--) {
			dummy = genrand_int32();
		}
		QueryPerformanceCounter((LARGE_INTEGER*)&endOriginal);

		// Implementation
		QueryPerformanceCounter((LARGE_INTEGER*)&start);
		for (size_t i = MaxIterations; i > 0; i--) {
			dummy = mersenneTwister32.next();
		}
		QueryPerformanceCounter((LARGE_INTEGER*)&end);

		Console::WriteLine("\n----- 32-bit test -----");
		Console::WriteLine("Time: {0} seconds.", ((end - start) * 1.0 / frequency).ToString());
		Console::WriteLine("Original time: {0} seconds.", ((endOriginal - startOriginal) * 1.0 / frequency).ToString());
		Console::WriteLine("Speed-up to original: {0} %.", (((endOriginal - startOriginal) * 1.0 / (end - start) * 1.0) * 100.0).ToString());
	}
	else {
		DWORD dwError = GetLastError();
		Console::WriteLine("Error value = {0}", dwError.ToString());
	}

	Console::WriteLine();
	Console::Write("Press ENTER to finish.");
	Console::Read();

	return 0;
}
